
typedef enum {
	VALUE_BYTE		= 0x00,	// (none; must be 0) ubyte[1] signed one-byte integer value 
	VALUE_SHORT		= 0x02,	// size - 1 (01) ubyte[size] signed two-byte integer value, sign-extended 
	VALUE_CHAR		= 0x03,	// size - 1 (01) ubyte[size] unsigned two-byte integer value, zero-extended 
	VALUE_INT		= 0x04,	// size - 1 (03) ubyte[size] signed four-byte integer value, sign-extended 
	VALUE_LONG		= 0x06, // size - 1 (07) ubyte[size] signed eight-byte integer value, sign-extended 
	VALUE_FLOAT		= 0x10, // size - 1 (03) ubyte[size] four-byte bit pattern, zero-extended to the right, and interpreted as an IEEE754 32-bit floating point value  
	VALUE_DOUBLE	= 0x11,	// size - 1 (07) ubyte[size] eight-byte bit pattern, zero-extended to the right, and interpreted as an IEEE754 64-bit floating point value  
	VALUE_STRING	= 0x17, // size - 1 (03) ubyte[size] unsigned (zero-extended) four-byte integer value, interpreted as an index into the string_ids section and representing a string value  
	VALUE_TYPE		= 0x18, // size - 1 (03) ubyte[size] unsigned (zero-extended) four-byte integer value, interpreted as an index into the type_ids section and representing a reflective type/class value  
	VALUE_FIELD     = 0x19, // size - 1 (03) ubyte[size] unsigned (zero-extended) four-byte integer value, interpreted as an index into the field_ids section and representing a reflective field value  
	VALUE_METHOD	= 0x1a, // size - 1 (03) ubyte[size] unsigned (zero-extended) four-byte integer value, interpreted as an index into the method_ids section and representing a reflective method value  
	VALUE_ENUM		= 0x1b, // size - 1 (03) ubyte[size] unsigned (zero-extended) four-byte integer value, interpreted as an index into the field_ids section and representing the value of an enumerated type constant  
	VALUE_ARRAY		= 0x1c, // (none; must be 0) encoded_array an array of values, in the format specified by "encoded_array Format" below. The size of the value is implicit in the encoding.  
	VALUE_ANNOTATION= 0x1d, // (none; must be 0) encoded_annotation a sub-annotation, in the format specified by "encoded_annotation Format" below. The size of the value is implicit in the encoding.  
	VALUE_NULL      = 0x1e,	// (none; must be 0) (none) null reference value 
	VALUE_BOOLEAN	= 0x1f
} annotation_valueE;

struct annotation_arrayS;
struct annotationS;

typedef union {
	s1T				_s1;
	s2T				_s2;
	u2T				_u2;
	s4T				_s4;
	u4T				_u4;
	s8T				_s8;
	u8T				_u8;
	float			_float;
	double			_double;
	stringT			*_stringP;
	fieldT			*_fieldP;
	methodT			*_methodP;
	struct annotation_arrayS	*_arrayP;
	struct annotationS			*_annotationP;
} valueT;

typedef struct {
	annotation_valueE	type;
	valueT				value;
} annotation_valueT;

typedef struct annotation_arrayS {
	u4T					size;
	annotation_valueT	value[1];
} annotation_arrayT;

typedef struct {
	// element name, represented as an index into the string_ids section. The string must conform to the syntax for MemberName, defined above.
	stringT				*nameP;
	annotation_valueT	*valueP;
} annotation_elementT;

typedef struct annotationS {
	// type of the annotation. This must be a class (not array or primitive) type.  
	stringT				*typeP;
	// number of name-value mappings in this annotation 
	u4T					size;
	// elements of the annotataion, represented directly in-line (not as offsets). Elements must be sorted in increasing order by string_id index.  
	annotation_elementT	*elementsP;
} annotationT;

typedef struct {
	// intended visibility of this annotation (see below) 
	u4T			visibility;
	// encoded annotation contents, in the format described by "encoded_annotation Format" under "encoded_value Encoding" above.
	annotationT	annotation;
} annotation_itemT;

typedef struct {
	u4T					size;
	annotation_itemT	*itemsP;
} annotation_setT;

typedef struct {
	u4T					method;
	u4T					size;
	annotation_setT		list[1];	// Variable length
} annotation_parameterT;

typedef struct {
	// offset from the start of the file to the annotations made directly on the class, or 0 if the class has no direct annotations. 
	// The offset, if non-zero, should be to a location in the data section. 
	// The format of the data is specified by "annotation_set_item" above.
	annotation_setT			class_annotations;
	// count of fields annotated by this item 
	u4T						fields;
	annotation_setT			*fieldsP;
	// count of methods annotated by this item 
	u4T						methods;
	annotation_setT			*methodsP;
	// count of method parameter lists annotated by this item 
	u4T						parameters;
	annotation_parameterT	**parametersPP;
} annotationsT;

class Cdirectory;
class Cfunction;

typedef struct {
	fieldT	*fieldP;
	u4T		access_flags;
} encoded_fieldT;

class Csource;
class Cfile;

class Cclass : public Cobject {
public:
	static Cclass	*g_class_hash[HASHSIZE];

	Cclass			*m_hashP;
	Cclass			*m_nextP;			// Linked list within Csource

	Csource			*m_sourceP;			// The source that contains this class
	stringT			*m_nameP;			// Class type of this class

	int				m_hypertext_cnt;

	u4T				m_access_flags;		// access flags for the class (public, final, etc.). See "access_flags Definitions" for details.
	stringT			*m_superclassP;		// type of the superclass or empty if no superclass. If present, this must be a class type, and not an array or primitive type.  
	u4T				m_interfaces;
	stringT			**m_interfacesPP;	// Interfaces this class supports
	// the file containing the original source for (at least most of) this class, or the special value NO_INDEX to represent a lack of this information. The debug_info_item of any given method may override this source file, but the expectation is that most classes will only come from one source file.  
	Cfile			*m_source_fileP;
	// the number of static fields defined in this item 
	u4T				m_static_fields;
	// the defined static fields, represented as a sequence of encoded elements. The fields must be sorted by field_idx in increasing order.  
	encoded_fieldT	*m_static_fieldsP;
	// the number of instance fields defined in this item 
	u4T				m_instance_fields;
	// the defined instance fields, represented as a sequence of encoded elements. The fields must be sorted by field_idx in increasing order.  
	encoded_fieldT	*m_instance_fieldsP;
	// direct_methods_size uleb128 the number of direct methods defined in this item 
	u4T				m_direct_functions;
	// the number of virtual methods defined in this item 
	u4T				m_virtual_functions;
	// the defined direct (any of static, private, or constructor) methods, represented as a sequence of encoded elements. The methods must be sorted by method_idx in increasing order.  
	Cfunction		*m_functionsP;
	// annotations_off uint offset from the start of the file to the annotations structure for this class, or 0 if there are no annotations on this class. This offset, if non-zero, should be in the data section, and the data there should be in the format specified by "annotations_directory_item" below, with all items referring to this class as the definer.  
	annotationsT	m_annotations;
	// static_values_off uint offset from the start of the file to the list of initial values for static fields, or 0 if there are none (and all static fields are to be initialized with 0 or null). This offset should be in the data section, and the data there should be in the format specified by "encoded_array_item" below. The size of the array must be no larger than the number of static fields declared by this class, and the elements correspond to the static fields in the same order as declared in the corresponding field_list. The type of each array element must match the declared type of its corresponding field. If there are fewer elements in the array than there are static fields, then the leftover fields are initialized with a type-appropriate 0 or null.  
	annotation_arrayT	*m_initial_valuesP;

	int				m_has_clones;			// 0 if no clones under this

	static u4T		hash(const u2T *nameP);
	static Cclass	*locate(stringT *nameP);
//	static Class	*locate_mangled(stringT *nameP);

	Cclass(void);
	
	void			emit_ta(void);


};