#include "typedef.h"

#include <memory.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>

#if defined(WIN32)
#include <io.h>
#include <direct.h>

#define open _open
#define read _read
#define close _close
#define lseek _lseek
#define wopen _wopen
#else
#include <wchar.h>
#include <errno.h>
#include <unistd.h>

#define _O_BINARY 0
#endif

#include "object.h"
#include "xmalloc.h"
#include "file.h"

Cfile	*Cfile::g_headP;
Cfile	**Cfile::g_tailPP = &g_headP;

Cfile::Cfile(void)
{
	static int number = -1;

	memset(this, 0, sizeof(Cfile));
	m_number  = ++number;
	*g_tailPP = this;
	g_tailPP  = &m_nextP;
}

/* This exists to permit file sharing of source files */

/* static */ Cfile *
Cfile::locate(const stringT *stringP)
{
	Cfile *fileP;

	// TODO improve efficiency
	for (fileP = g_headP; fileP; fileP = fileP->m_nextP) {
		if (stringP->P == fileP->m_nameP->P) {
			return(fileP);
	}	}

	fileP = new Cfile();
	fileP->m_nameP = stringP;
	return fileP;
}

#ifdef WIN32
#define uopen(A, B, C) _wopen((wchar_t *) A, B, C)
#else
// Don't know how to do a wopen on linux
#include <stdlib.h>

int
uopen(u2T *pathP, int options, int permissions)
{
	char	*path1P;
	u2T		*P;
	char	*P1;

	P1 = path1P;
	for (P = pathP; *P++; );
	path1P = (char *) alloca(P - pathP);

	for (P = pathP; (*P1++ = *P++); );
	return(open(path1P, options, permissions));
}
#endif

void
Cfile::load_source(void)
{
	extern const char	*g_sourcepathP;

	const char	*sourcepathP;	
	u2T		*filenameP = m_nameP->P;
	u2T		path[2048];					// N.B. These are 4 byte chars on UNIX
	u2T		*pathP;
	u2T		*P1;
	const char	*P;
	int		c;

	int		fileno;
	int		size, ret, lineno;
	char	*source_codeP, *atP, *eolP;
	char	**sourcePP;

	sourcepathP = g_sourcepathP;
	if (!sourcepathP) {
		sourcepathP = ".";
	}
	pathP  = path;
	for (P = sourcepathP;; ++P) {
		switch (c = *P) {
		case 0:
		case ';':
		case ',':
			*pathP++ = '/';
			break;
		default:
			*pathP++ = (u2T) c;
			continue;
		}
		for (P1 = filenameP; (*pathP++ = *P1++); );	// Potentially convert character sizes
		fileno = uopen(path, O_RDONLY | _O_BINARY, 0);
		if (fileno < 0) {
			if (!c) {
				break;
			}
			pathP = path;
			continue;
		}
		size = lseek(fileno, 0L, SEEK_END);
		if (size > 0) {
			if (lseek(fileno, 0L, SEEK_SET) < 0) {
				ret = -1;
			} else {
				source_codeP = (char *) Xmalloc(size + 2);
				if (!source_codeP) {
					outofmemory();
				}
				ret = read(fileno, source_codeP, size);
			}
			close(fileno);
			if (ret == size) {
				atP         = source_codeP;
				if (size == 0 || atP[size-1] != '\n') {
					atP[size]   = '\n';
					atP[size+1] = 0;
				} else {
					atP[size]   = 0;
				}

				lineno      = 0;
				for (; (eolP = strchr(atP, '\n')); atP = eolP + 1) {
					++lineno;
				}
				m_sourcePP     = sourcePP = (char **) Xmalloc((lineno+1) * sizeof(char *));
				m_source_lines = lineno;

				lineno         = 0;
				for (atP = source_codeP; (eolP = strchr(atP, '\n')); atP = eolP + 1) {
					sourcePP[lineno++] = atP;
					*eolP              = 0;
				}
				sourcePP[lineno]       = atP;
				return;
	}	}	}
	m_sourcePP = ((char **) -1);
}

extern FILE	*stdoutF;

void
Cfile::showline(int lineno) const
{
	if (lineno > 0 && lineno <= m_source_lines) {
		char	*atP = m_sourcePP[lineno-1];
		char	c;

		fprintf(stdoutF, "<i>%d: ", lineno);
		while ((c = *atP++) != '\n' && c) {
			switch (c) {
			case '\r':
				continue;
			case '>':
				fputs("&gt;", stdoutF);
				continue;
			case '<':
				fputs("&lt;", stdoutF);
				continue;
			case '&':
				fputs("&amp;", stdoutF);
				continue;
			case '\t':
				c = ' ';
			}
			fputc(c, stdoutF);
		}
		fputs("</i>", stdoutF);
	}
}

