#include "typedef.h"

#include <assert.h>
//#include <sys/types.h>
//#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>

#ifdef WIN32
#include <io.h>
#include <direct.h>
#include <sys/types.h>
#include <sys/stat.h>

#define open _open
#define read _read
#define close _close
#define alloca _alloca
#define getcwd _getcwd
#define lseek _lseek
#define access _access
#define stat _stat

#define F_OK 0
#else
#include <wchar.h>
#include <errno.h>
#include <unistd.h>
#include <alloca.h>

#define _O_BINARY 0
#endif

#include "stringcache.h"
#include "util.h"
#include "xmalloc.h"
#include "object.h"
#include "opcode.h"
#include "code.h"
#include "source.h"
#include "class.h"
#include "function.h"
#include "dex.h"

extern const char	*g_classpathP;
extern int	g_read;

extern FILE	*stdoutF;
u1T			*g_imageP         = 0;
int			g_image_size      = 0;
static int	g_image_max       = -1;

/* Read the class image from disk */

static int
read_image(const char *filenameP)
{
	int		fileno;
	int		ret;

	fileno = open(filenameP, O_RDONLY | _O_BINARY, 0);
	if (fileno == -1) {
		return(1);
	}
	g_image_size = lseek(fileno, 0L, SEEK_END);
	if (g_image_size < 0) {
		fprintf(stderr, "Can't lseek end of %s\n", filenameP);
		goto fail;
	}
	if (lseek(fileno, 0L, SEEK_SET) < 0) {
		fprintf(stderr, "Can't lseek start of %s\n", filenameP);
		goto fail;
	}
	if (g_image_size > g_image_max) {
		if (g_imageP) {
			Xfree(g_imageP);
		}
		g_imageP     = (u1T *) Xmalloc(g_image_size + 1);
		if (!g_imageP) {
			outofmemory();
		}
		g_image_max  = g_image_size;
	}

	ret = read(fileno, g_imageP, g_image_size);
	if (ret != g_image_size) {
		fprintf(stderr, "Can't read %s\n", filenameP);
		goto fail;
	}

	g_imageP[g_image_size] = 0;
	ret = 0;
done:
	if (close(fileno)) {
		fprintf(stderr, "Can't close %s\n", filenameP);
		ret = -1;
	}
	return(ret);
fail:
	ret = -1;
	goto done;
}

extern int load_dex_file(void);

int
readclassfile(const char *pathP, Cclass *classP)
{
	Csource			*sourceP;
	int				ret;

	if (classP) {
		if (access(pathP, F_OK)) {
			return(1);
	}	}

	ret = read_image(pathP);
	if (ret != 0) {
		if (ret == 1 && !classP) {
			fprintf(stderr, "Can't open %s\n", pathP);
		}
		goto done;
	}

	++g_read;

	sourceP = new Csource();
	ret = sourceP->load(pathP, (classP != 0));
	if (ret != 0) {
		goto done;
	}


done:
	return(ret);
}

void
pseudo_source(void)
{
	int			i, ret;
	Cclass		*classP;
	const char	*classpathP;
	int			repeat;
	char		path[2048];
	const char	*P;
	char		*pathP;
	u2T			*P2;
	char		c;

	classpathP = g_classpathP;
	if (!classpathP) {
		classpathP = ".";
	}
	do {
		repeat = 0;
		for (i = 0; i < HASHSIZE; ++i) {
			for (classP = Cclass::g_class_hash[i]; classP; classP = classP->m_hashP) {

				if (!classP->m_sourceP) {
					pathP = path;
					for (P = classpathP;; ++P) {
						switch (c = *P) {
						case 0:
						case ';':
						case ',':
							*pathP++ = '/';
							break;
						default:
							*pathP++ = c;
							continue;
						}
						P2 = classP->m_nameP->P;
						assert(*P2 == 'L');
						for (++P2; (*pathP++ = (u1T) *P2++); );
						--pathP;
						if (pathP[-1] == ';') {
							--pathP;
						}
						strcpy(pathP, ".dex");
						ret = readclassfile(path, classP);
						switch (ret) {
						case 0:
							// Could read
							repeat = 1;
							assert(classP->m_sourceP);
							break;
						case 1:
							if (c && P[1]) {
								// Wrong item on class path used; move to next item on class path
								pathP = path;
								continue;
							}
							fputs("<font color=red>Can't locate class ", stdoutF);
							Cfunction::dump_descriptor(classP->m_nameP->P);
							fputs("</font>\n<br>\n", stdoutF);
							classP->m_sourceP = (Csource *) -1;
							break;
						}
						break;
		}	}	}	}
	} while (repeat);
}

