#include "typedef.h"

#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <search.h>

#include "xmalloc.h"
#include "util.h"
#include "object.h"
#include "opcode.h"
#include "code.h"
#include "source.h"
#include "sort.h"

int		Csort::g_labels_count = 0;
int		Csort::g_labels_max   = 0;
labelT	*Csort::g_labelsP     = 0;	

void	
Csort::clear_sort(void)
{
	g_labels_count = 0;
}

void
Csort::append_exception(u4T from, u4T to, u4T order, u4T classtype, u4T handler)
{
	labelT	*labelP;

	if (g_labels_max <= g_labels_count) {
		if (!g_labels_count) {
			g_labelsP    = (labelT *) Xmalloc(1024 * sizeof(labelT));
			g_labels_max = 1024;	// Must be even
		} else {
			g_labelsP      = (labelT *) Xrealloc(g_labelsP, (g_labels_max << 1));
			g_labels_max <<= 1;
	}	}

	labelP            = g_labelsP + g_labels_count++;
	labelP->from      = from;
	labelP->to        = to;
	labelP->order     = order;
	labelP->classtype = classtype;
	labelP->handler   = handler;
	labelP->enabled   = 1;

	labelP            = g_labelsP + g_labels_count++;
	labelP->from      = to;
	labelP->to        = from;
	labelP->order     = order;
	labelP->classtype = classtype;
	labelP->handler   = handler;
	labelP->enabled   = 0;
}

/*
	< 0 P1 less than P2
      0 P1 equals    P2
    > 0 P1 greater   P2
 */

static int
label_compare(const void *P1, const void *P2)
{
	labelT	*label1P, *label2P;
	int		diff;

	label1P = (labelT *) P1;
	label2P = (labelT *) P2;
	diff    = label1P->from - label2P->from;

	if (diff) {
		// label1 occurs before label2 if label1 from is smaller
		return (diff);
	}

	diff = label1P->enabled - label2P->enabled;
	if (diff) {
		// label1P occurs before label2P if label1P is not enabled and label2P is
		return(diff);
	}

	// If enabled label1P occurs before label2P if label1P->to > label2P->to
	// If not enabled label1P occurs before label2P if label1P->to > label2P->to
	diff = label2P->to - label1P->to;
	if (diff) {
		return(diff);
	}

	diff = label1P->order - label2P->order;
	if (!label1P->enabled) {
		diff = -diff;
	}
	return(diff);
}

void
Csort::sort(void)
{
	qsort(g_labelsP, g_labels_count, sizeof(labelT), label_compare);
}

